import argparse
import json
import sys
import typing as t

import randog.factory

from ._utils.type import positive_int


class Args:
    _args: argparse.Namespace

    def __init__(self, argv: t.Sequence[str]):
        parser = argparse.ArgumentParser(
            prog="randog",
            usage="python -m randog [options] FACTORY_PATH [FACTORY_PATH ...]",
            description="Create object at random.",
        )
        group_output_fmt = parser.add_mutually_exclusive_group()
        parser.add_argument(
            "factories",
            nargs="+",
            metavar="FACTORY_PATH",
            help="path of factory definition files",
        )
        parser.add_argument(
            "--repeat",
            "-r",
            metavar="COUNT",
            default=1,
            type=positive_int,
            help=(
                "repeat generation a specified number of times. "
                "The results are output one by one; if you want them as a single list, use --list instead."
            ),
        )
        parser.add_argument(
            "--list",
            "-L",
            metavar="LENGTH",
            type=positive_int,
            help=(
                "if specified, repeats the specified numerical generation "
                "and returns a list consisting of the results."
            ),
        )
        group_output_fmt.add_argument(
            "--repr",
            dest="output_fmt",
            action="store_const",
            const="repr",
            help="if specified, it outputs generated object by repr()",
        )
        group_output_fmt.add_argument(
            "--json",
            dest="output_fmt",
            action="store_const",
            const="json",
            help="if specified, it outputs generated object by json format",
        )
        parser.add_argument(
            "--output",
            "-O",
            metavar="DESC_PATH",
            help="destination file path",
        )
        self._args = parser.parse_args(argv[1:])

    @property
    def factories(self) -> t.Sequence[str]:
        return self._args.factories

    @property
    def repeat(self) -> int:
        return self._args.repeat

    @property
    def list(self) -> t.Optional[int]:
        return self._args.list

    @property
    def output_fmt(self) -> t.Optional[str]:
        return self._args.output_fmt

    @property
    def output_path(self) -> t.Optional[str]:
        return self._args.output

    @property
    def multiple_output_path(self) -> bool:
        if self._args.output is None:
            return False
        else:
            return self.output_path_for(1) != self.output_path_for(2)

    def output_path_for(self, number: int) -> t.Optional[str]:
        if self._args.output is None:
            return None
        else:
            return self._args.output.format(number)


def _build_factories(args: Args) -> t.Iterator[randog.factory.Factory]:
    for filepath in args.factories:
        if filepath == "-":
            yield randog.factory.from_pyfile(sys.stdin)
        else:
            yield randog.factory.from_pyfile(filepath)


class _DummyIO:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def _open_output_fp_only(args: Args) -> t.Union[_DummyIO, t.TextIO]:
    if args.output_path is None or args.multiple_output_path:
        return _DummyIO()
    else:
        return open(args.output_path, mode="wt")


def _open_output_fp_numbered(args: Args, number: int) -> t.Union[_DummyIO, t.TextIO]:
    if args.output_path is None or not args.multiple_output_path:
        return _DummyIO()
    else:
        return open(args.output_path_for(number), mode="wt")


def _output_generated(generated: t.Any, fp: t.TextIO, args: Args):
    if args.output_fmt == "repr":
        print(repr(generated), file=fp)
    elif args.output_fmt == "json":
        json.dump(generated, fp, default=str)
        fp.write("\n")
    else:
        print(generated, file=fp)


def main():
    args = Args(sys.argv)

    with _open_output_fp_only(args) as fp_only:
        index = 0
        for factory in _build_factories(args):
            for r_index in range(args.repeat):
                if args.list is None:
                    generated = factory.next()
                else:
                    generated = list(factory.iter(args.list))

                with _open_output_fp_numbered(args, index) as fp_numbered:
                    if not isinstance(fp_numbered, _DummyIO):
                        fp = fp_numbered
                    elif not isinstance(fp_only, _DummyIO):
                        fp = fp_only
                    else:
                        fp = sys.stdout

                    _output_generated(generated, fp, args=args)

                index += 1


if __name__ == "__main__":
    main()



from os import PathLike
import types
import typing as t

from randog.factory import Factory


def from_pyfile(file: t.Union[str, PathLike, t.IO]) -> Factory:
    """Returns a factory defined in the specified file.

    Parameters
    ----------
    file : str | PathLike | IO
        the filename of the factory definition
    """
    if isinstance(file, (str, PathLike)):
        with open(file, mode="rb") as fp:
            return _from_pyfile(fp, file)
    else:
        return _from_pyfile(file, "<io>")


def _from_pyfile(fp: t.IO, filename: t.Union[str, PathLike]) -> Factory:
    import randog

    d = types.ModuleType("factory")
    d.randog = randog
    d.__file__ = filename
    try:
        exec(compile(fp.read(), filename, "exec"), d.__dict__)
    except OSError as e:
        e.strerror = f"Unable to load factory file ({e.strerror})"
        raise

    attr_name = "FACTORY"

    if not hasattr(d, attr_name):
        raise AttributeError(
            f"factory file '{filename}' has no attribute '{attr_name}'"
        )

    factory = getattr(d, attr_name)

    if not isinstance(factory, Factory):
        raise AttributeError(
            f"attribute '{attr_name}' of factory file '{filename}' is not a factory"
        )

    return factory



import datetime
from decimal import Decimal
import typing as t

import randog.factory
from randog.exceptions import FactoryConstructionError
from .._utils.nullsafe import dfor


def custom(example, **kwargs):
    """custom_func for `randog.factory.from_example` to generate sqlalchemy-derived objects

    This function is an add-on for `randog.factory.from_example`.

    Specify this function as `custom_func` if you want to create a factory that generates sqlalchemy-derived objects in
    `from_example`.

    Examples
    --------
    >>> from sqlalchemy import Column, Integer, String
    >>> from sqlalchemy.orm import declarative_base
    >>> import randog.factory
    >>> import randog.sqlalchemy
    >>>
    >>> Base = declarative_base()
    >>>
    >>> class User(Base):
    ...     __tablename__ = "user"
    ...
    ...     id = Column(Integer, primary_key=True, autoincrement=True)
    ...     name = Column(String)
    >>>
    >>> # specify `randog.sqlalchemy.custom` as `custom_func`
    >>> factory = randog.factory.from_example(User, custom_func=randog.sqlalchemy.custom)
    >>> generated = factory.next()
    """
    if type(example).__module__.startswith("sqlalchemy."):
        # example is Model
        if hasattr(example, "_sa_class_manager"):
            return _custom_func_for_model(example._sa_class_manager, **kwargs)
        # example is Column or like it
        else:
            return _custom_func_for_column(example, **kwargs)
    else:
        return NotImplemented


def _custom_func_for_model(
    class_manager,
    **kwargs,
):
    return dict(class_manager)


def _custom_func_for_column(
    column,  # : sqlalchemy.sql.ColumnElement,
    *,
    context: randog.factory.FromExampleContext,
    **kwargs,
):
    # type checks of `column`
    column_type, column_python_type = _get_column_types(column)
    enums = getattr(column_type, "enums", None) if column_type is not None else None
    autoincrement = getattr(column, "autoincrement", False)

    # normalize example for the function `from_example`.
    # If this is difficult, create the factory directly.
    factory = None
    customized_example = None
    if enums is not None and column_python_type == str:
        # use randchoice instead from_example(str) if it is enumeration
        factory = randog.factory.randchoice(*enums, rnd=context.rnd)
    elif autoincrement and column_python_type == int:
        # use increment instead from_example(int) if it is auto incremental
        factory = randog.factory.increment(maximum=2**31 - 1, rnd=context.rnd)
    elif column_python_type in __CUSTOM_FUNC_FOR_SPEC_TYPE:
        customized_example = __CUSTOM_FUNC_FOR_SPEC_TYPE[column_python_type](
            column_type, context=context, **kwargs
        )
    elif isinstance(column_python_type, type):
        customized_example = column_python_type
    else:
        raise FactoryConstructionError(
            f"cannot create factory from sqlalchemy.Column type of {column_type}: not implemented"
        )

    # Create a factory based on the normalized example.
    # If a factory has already been created directly, it is used.
    if factory is None:
        factory = randog.factory.from_example(customized_example, context=context)
    # make the factory nullable
    if hasattr(column, "nullable") and column.nullable:
        factory = factory.or_none(rnd=context.rnd)

    return factory


def _custom_func_for_str_column(
    column_type,
    **kwargs,
):
    length = getattr(column_type, "length", None)
    if length is not None:
        return randog.factory.randstr(length=length)
    else:
        return str


def _custom_func_for_numeric_column(
    column_type,
    **kwargs,
):
    precision = getattr(column_type, "precision", None)
    scale = getattr(column_type, "scale", None)

    if precision is not None:
        scale_n = dfor(scale, 0)

        maximum = Decimal("1").scaleb(precision - scale_n) - Decimal("1").scaleb(
            -scale_n
        )
        minimum = -maximum
    else:
        minimum = None
        maximum = None

    return randog.factory.randdecimal(minimum, maximum, decimal_len=scale)


def _custom_func_for_datetime_column(
    column_type,
    **kwargs,
):
    use_timezone = getattr(column_type, "timezone", None)

    if use_timezone:
        return randog.factory.randdatetime(tzinfo=datetime.timezone.utc)
    else:
        return randog.factory.randdatetime(tzinfo=None)


def _custom_func_for_time_column(
    column_type,
    **kwargs,
):
    use_timezone = getattr(column_type, "timezone", None)

    if use_timezone:
        return randog.factory.randtime(tzinfo=datetime.timezone.utc)
    else:
        return randog.factory.randtime(tzinfo=None)


__CUSTOM_FUNC_FOR_SPEC_TYPE = {
    str: _custom_func_for_str_column,
    Decimal: _custom_func_for_numeric_column,
    datetime.datetime: _custom_func_for_datetime_column,
    datetime.time: _custom_func_for_time_column,
}


def _get_column_types(column) -> t.Tuple[t.Any, t.Any]:
    # type checks of `column`
    if hasattr(column, "type"):
        column_type = column.type
    else:
        raise FactoryConstructionError(
            f"cannot create factory from {column}: not implemented"
        )
    if hasattr(column_type, "python_type"):
        column_python_type = column_type.python_type
    else:
        raise FactoryConstructionError(
            f"cannot create factory from {column}: not implemented"
        )

    return column_type, column_python_type



from abc import ABC, abstractmethod
import typing as t
from random import Random

T = t.TypeVar("T")
R = t.TypeVar("R")


class Factory(ABC, t.Generic[T]):
    @abstractmethod
    def next(self) -> T:
        """Generate a value randomly according to the rules specified when assembling the factory.

        Returns
        -------
        T
            a value generated randomly
        """
        pass

    def or_none(
        self,
        prob: float = 0.1,
        *,
        rnd: t.Optional[Random] = None,
    ) -> "Factory[t.Union[T, None]]":
        """Returns a factory whose result may be None with the specified probability.

        Examples
        --------
        >>> import randog
        >>>
        >>> factory = randog.factory.randstr().or_none(0.2)
        >>>
        >>> generated = factory.next()
        >>> assert generated is None or isinstance(generated, str)

        Parameters
        ----------
        prob : float, default=0.1
            Probability that the result is None
        rnd : Random, optional
            random number generator to be used

        Returns
        -------
        Factory[T|None]
            A factory whose result may be None with the specified probability.
        """
        import randog.factory

        return randog.factory.union(
            self,
            randog.factory.const(None),
            weights=[1 - prob, prob],
            rnd=rnd,
        )

    def post_process(self, post_process: t.Callable[[T], R]) -> "Factory[R]":
        """Returns a factory whose result will be modified by `post_process`

        Examples
        --------
        >>> import randog
        >>>
        >>> # use post_process to format the random decimal value
        >>> factory = (
        ...     randog.factory.randdecimal(0, 50000, decimal_len=2)
        ...                  .post_process(lambda x: f"${x:,}")
        ... )
        >>>
        >>> # examples: '$12,345.67', '$3,153.21', '$12.90', etc.
        >>> generated = factory.next()
        >>> assert isinstance(generated, str)
        >>> assert generated[0] == "$"

        Parameters
        ----------
        post_process : Callable[[T], R]
            the mapping to modify the result

        Returns
        -------
        Factory[R]
            A factory whose result will be modified by `post_process`.
        """
        return PostFactory(self, post_process)

    def iter(self, size: int) -> t.Iterator[T]:
        """Returns an iterator which serves result randomly `size` times.

        Examples
        --------
        >>> import randog
        >>> factory = randog.factory.randstr(length=5)
        >>>
        >>> for result in factory.iter(10):
        ...     assert isinstance(result, str)
        >>>
        >>> results = list(factory.iter(5))
        >>> assert len(results) == 5

        Parameters
        ----------
        size : int
            the number of the iterator

        Returns
        -------
        Iterator[T]
            An iterator
        """
        for _ in range(size):
            yield self.next()

    def infinity_iter(self) -> t.Iterator[T]:
        """Returns an infinity iterator which serves result randomly.

        The result is INFINITY so do NOT use it directly with `for`, `list`, and so on.

        Examples
        --------
        >>> import randog
        >>> factory = randog.factory.randstr(length=5)
        >>>
        >>> keys = ["foo", "bar"]
        >>> for k, v in zip(keys, factory.infinity_iter()):
        ...     assert k in keys
        ...     assert isinstance(v, str)

        Returns
        -------
        Iterator[T]
            An infinity iterator
        """
        while True:
            yield self.next()


class PostFactory(Factory[R], t.Generic[T, R]):
    _base_factory: Factory[T]
    _post_process: t.Callable[[T], R]

    def __init__(self, base_factory: Factory[T], post_process: t.Callable[[T], R]):
        self._base_factory = base_factory
        self._post_process = post_process

    def next(self) -> R:
        pre_result = self._base_factory.next()
        return self._post_process(pre_result)



import datetime as dt
import typing as t
from random import Random

from ._base import Factory
from .._utils.nullsafe import dfor


def randtime(
    *,
    tzinfo: t.Optional[dt.tzinfo] = None,
    rnd: t.Optional[Random] = None,
) -> Factory[dt.time]:
    """Return a factory generating random time values.

    Parameters
    ----------
    tzinfo : tzinfo, optional
        the timezone information
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    return TimeRandomFactory(
        tzinfo=tzinfo,
        rnd=rnd,
    )


class TimeRandomFactory(Factory[dt.time]):
    """factory generating random time values"""

    _random: Random
    _tzinfo: dt.tzinfo

    def __init__(
        self,
        *,
        tzinfo: t.Optional[dt.tzinfo] = None,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random time values.

        Parameters
        ----------
        tzinfo : tzinfo, optional
            the timezone information
        rnd : Random, optional
            random number generator to be used
        """
        self._random = dfor(rnd, Random())
        self._tzinfo = tzinfo

    def next(self) -> dt.time:
        hours = self._random.randint(0, 23)
        minutes = self._random.randint(0, 59)
        seconds = self._random.randint(0, 59)
        microseconds = self._random.randint(0, 999_999)

        return dt.time(
            hours,
            minutes,
            seconds,
            microseconds,
            tzinfo=self._tzinfo,
        )
